import pytest
from django.db import IntegrityError

from pik.core.tests.fixtures import create_user
from ..models import MyOwned, MyNullOwned
from .factories import MyOwnedFactory, MyNullOwnedFactory


@pytest.fixture(params=[
    (MyOwned, MyOwnedFactory),
])
def owned_model(request):
    return request.param


@pytest.fixture(params=[
    (MyNullOwned, MyNullOwnedFactory),
])
def null_owned_model(request):
    return request.param


def test_owned_protocol(owned_model):
    model, factory = owned_model
    user = create_user()

    obj = factory.create(user=user)
    assert obj.user_id


def test_owned_protocol_no_user_problem(owned_model):
    model, factory = owned_model
    with pytest.raises(IntegrityError):
        factory.create(user=None)


def test_null_owned_protocol(null_owned_model):
    model, factory = null_owned_model
    user = create_user()

    obj = factory.create(user=user)
    assert obj.user_id


def test_null_owned_protocol_on_user_ok(null_owned_model):
    model, factory = null_owned_model
    obj = factory.create(user=None)
    assert obj.user is None



import logging
from typing import Optional, Type, Tuple, Union, List

from django.core.exceptions import ValidationError
from django.db import models, IntegrityError

LOGGER = logging.getLogger(__name__)


def _update_m2m_fields(_obj, **kwargs):
    for key, values in kwargs.items():
        for value in values:
            m2m_set = getattr(_obj, key)
            if value not in m2m_set.all():
                m2m_set.add(value)


def _get_m2m_kwargs(_model, **kwargs):
    m2m_kwargs = {}
    for field in _model._meta.get_fields():  # noqa: pylint=protected-access
        if (field.get_internal_type() == 'ManyToManyField'
                and field.name in kwargs):
            m2m_kwargs[field.name] = kwargs.pop(field.name)
    return m2m_kwargs, kwargs


def get_object_or_none(
        source: Union[Type[models.Model], models.QuerySet, models.Manager],
        *args, **kwargs) -> Optional[models.Model]:
    assert (isinstance(source, (models.QuerySet, models.Manager))
            or issubclass(source, models.Model))

    if not isinstance(source, (models.QuerySet, models.Manager)):
        source = source.objects

    try:
        return source.get(*args, **kwargs)
    except source.model.DoesNotExist:
        return None


def validate_and_create_object(model: Type[models.Model], **kwargs) \
        -> models.Model:
    """
    :raises ValueError
    :return obj
    """
    assert issubclass(model, models.Model)

    m2m_kwargs, kwargs = _get_m2m_kwargs(model, **kwargs)

    obj = model(**kwargs)
    try:
        obj.full_clean()
        obj.save()
        if m2m_kwargs:
            _update_m2m_fields(obj, **m2m_kwargs)

    except (ValidationError, IntegrityError) as exc:
        LOGGER.warning(
            'Create %s error: %r (kwargs=%r)', model.__name__, exc, kwargs)
        raise ValueError(str(exc))
    return obj


def validate_and_update_object(obj: models.Model, **kwargs) \
        -> Tuple[models.Model, List[str]]:
    """
    :raises ValueError
    :return obj, is_updated
    """
    assert isinstance(obj, models.Model)
    model = type(obj)

    m2m_kwargs, kwargs = _get_m2m_kwargs(model, **kwargs)

    updated_keys = {}
    for key, value in kwargs.items():
        old_value = getattr(obj, key)
        if old_value == value:
            continue
        setattr(obj, key, value)
        updated_keys[key] = old_value

    if updated_keys:
        try:
            obj.full_clean()
            obj.save()
            if m2m_kwargs:
                _update_m2m_fields(obj, **m2m_kwargs)

        except (ValidationError, IntegrityError) as exc:
            for key, old_value in updated_keys.items():
                setattr(obj, key, old_value)
            LOGGER.warning(
                'Update %s error: %r (kwargs=%r)', model.__name__, exc, kwargs)
            raise ValueError(str(exc))
    return obj, list(updated_keys.keys())


def update_or_create_object(
        model: Type[models.Model],
        search_keys: Optional[dict] = None,
        **kwargs) \
        -> Tuple[models.Model, List[str], bool]:
    """
    :raises ValueError
    :return obj, is_updated, is_created
    """
    obj = get_object_or_none(model, **search_keys) if search_keys else None
    if obj:
        is_created = False
        obj, updates = validate_and_update_object(obj, **kwargs)
    else:
        updates = []
        is_created = True
        obj = validate_and_create_object(model, **kwargs)
    return obj, updates, is_created



import factory

from ..models import MyDated, MyOwned, MyNullOwned, MyUided, MyPUided, \
    MyVersioned, MyStrictVersioned, MyHistorized, MyBasePHistorical, \
    MyBaseHistorical


class MyDatedFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MyDated


class MyOwnedFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MyOwned


class MyNullOwnedFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MyNullOwned


class MyUidedFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MyUided


class MyPUidedFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MyPUided


class MyVersionedFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MyVersioned


class MyStrictVersionedFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MyStrictVersioned


class MyHistorizedFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MyHistorized


class MyBasePHistoricalFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MyBasePHistorical


class MyBaseHistoricalFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = MyBaseHistorical



from django.db import models

from pik.core.models import BasePHistorical


class TestNameModel(BasePHistorical):
    name = models.CharField(max_length=255)


class MySimpleModel(BasePHistorical):
    data = models.CharField(max_length=255)
    names = models.ManyToManyField(TestNameModel, blank=True)



import factory
from django.utils.crypto import get_random_string

from ..models import MySimpleModel, TestNameModel


class MySimpleModelFactory(factory.django.DjangoModelFactory):
    data = factory.LazyFunction(get_random_string)

    @factory.post_generation
    def names(self, create, extracted):
        if not create:
            return
        if extracted:
            for name in extracted:
                self.names.add(name)

    class Meta:
        model = MySimpleModel


class TestNameModelFactory(factory.django.DjangoModelFactory):
    name = factory.LazyFunction(get_random_string)

    class Meta:
        model = TestNameModel



from collections import Counter
from django.db.models.expressions import Col
from django.utils.timezone import now
from operator import attrgetter

from django.db import transaction
from django.db.models import signals, sql
from django.db.models.deletion import Collector
from django.db.models.fields.related import ForeignObject
from django.utils import six
from django.utils.translation import ugettext_lazy as _
from django.conf import settings

FIELD = 'deleted'


class DeleteNotSoftDeletedModel(Exception):
    pass


def _is_soft_excluded(model):
    soft_delete_exclude_list = getattr(settings, 'SOFT_DELETE_EXCLUDE', [])
    value = f'{model._meta.app_label}.{model._meta.object_name}'
    if value in soft_delete_exclude_list:
        return True
    return False


def _delete(self):
    from .soft_deleted import SoftDeleted
    safe_mode = getattr(settings, 'SOFT_DELETE_SAFE_MODE', True)

    time = now()

    # sort instance collections
    for model, instances in self.data.items():
        self.data[model] = sorted(instances, key=attrgetter("pk"))

    # if possible, bring the models in an order suitable for databases that
    # don't support transactions or cannot defer constraint checks until the
    # end of a transaction.
    self.sort()
    # number of objects deleted for each model label
    deleted_counter = Counter()

    with transaction.atomic(using=self.using, savepoint=False):
        # send pre_delete signals
        for model, obj in self.instances_with_model():
            if not model._meta.auto_created and not issubclass(  # noqa: pylint=protected-access
                    model, SoftDeleted):

                if safe_mode and not _is_soft_excluded(model):
                    raise DeleteNotSoftDeletedModel(
                        _(f'You are trying to delete {model._meta.object_name} instance,'
                          f' but {model._meta.object_name} is not subclass of '
                          f'{SoftDeleted._meta.object_name}. You need to inherit your '
                          f'model from {SoftDeleted._meta.object_name}'
                          f' or set settings.SOFT_DELETE_SAFE_MODE to False'))
                signals.pre_delete.send(
                    sender=model, instance=obj, using=self.using
                )

        # fast deletes
        for qs in self.fast_deletes:
            if issubclass(qs.model, SoftDeleted):
                pk_list = [obj.pk for obj in qs]
                qs = sql.UpdateQuery(qs.model)
                qs.update_batch(pk_list, {FIELD: time}, self.using)
                count = len(pk_list)
            else:
                count = qs._raw_delete(using=self.using)
            deleted_counter[qs.model._meta.label] += count

        # update fields
        for model, instances_for_fieldvalues in six.iteritems(self.field_updates):
            for (field, value), instances in six.iteritems(instances_for_fieldvalues):
                query = sql.UpdateQuery(model)
                query.update_batch([obj.pk for obj in instances],
                                   {field.name: value}, self.using)

        # reverse instance collections
        for instances in six.itervalues(self.data):
            instances.reverse()

        # delete instances
        for model, instances in six.iteritems(self.data):
            if issubclass(model, SoftDeleted):
                query = sql.UpdateQuery(model)
                pk_list = [obj.pk for obj in instances]
                query.update_batch(pk_list, {FIELD: time}, self.using)
                for instance in instances:
                    setattr(instance, FIELD, time)
                count = len(pk_list)
            else:
                query = sql.DeleteQuery(model)
                pk_list = [obj.pk for obj in instances]
                count = query.delete_batch(pk_list, self.using)

                if not model._meta.auto_created:
                    for obj in instances:
                        signals.post_delete.send(
                            sender=model, instance=obj, using=self.using
                        )
            deleted_counter[model._meta.label] += count

    # update collected instances
    for model, instances_for_fieldvalues in six.iteritems(self.field_updates):
        for (field, value), instances in six.iteritems(instances_for_fieldvalues):
            for obj in instances:
                setattr(obj, field.attname, value)
    for model, instances in six.iteritems(self.data):
        for instance in instances:
            setattr(instance, model._meta.pk.attname, None)
    return sum(deleted_counter.values()), dict(deleted_counter)


def get_extra_restriction_patch(func):
    def wrapper(self, where_class, alias, related_alias):
        cond = func(self, where_class, alias, related_alias)

        from .soft_deleted import SoftDeleted, _AllWhereNode
        if not issubclass(self.model, SoftDeleted) or issubclass(where_class, _AllWhereNode):
            return cond

        cond = cond or where_class()
        field = self.model._meta.get_field(FIELD)
        lookup = field.get_lookup('isnull')(field.get_col(related_alias), True)
        cond.add(lookup, 'AND')

        return cond
    return wrapper


# MONKEY PATCHES

# we want to prevent hard delete for SoftDeleted models
Collector.delete = _delete
# we want to remove objects from related QS
ForeignObject.get_extra_restriction = get_extra_restriction_patch(ForeignObject.get_extra_restriction)



from django.db import models

from pik.core.models import SoftDeleted, BasePHistorical


class _BaseBasePHistoricalTestModel(BasePHistorical):
    class Meta:
        abstract = True

    def __str__(self):
        return str(self.pk)


class MySoftDeleteModel(SoftDeleted, _BaseBasePHistoricalTestModel):
    name = models.CharField(max_length=255, blank=True, null=True)


class MyRelatedSoftDeletedModel(SoftDeleted, _BaseBasePHistoricalTestModel):
    name = models.CharField(max_length=255, blank=True, null=True)
    soft_deleted_fk = models.ForeignKey(
        MySoftDeleteModel, on_delete=models.CASCADE)


class MyRelatedNotSoftDeletedModel(_BaseBasePHistoricalTestModel):
    name = models.CharField(max_length=255, blank=True, null=True)
    soft_deleted_fk = models.ForeignKey(
        MySoftDeleteModel, on_delete=models.CASCADE)


class MyNotSoftDeletedModel(_BaseBasePHistoricalTestModel):
    name = models.CharField(max_length=255, blank=True, null=True)


class MySoftDeletedModelWithFK(SoftDeleted, _BaseBasePHistoricalTestModel):
    name = models.CharField(max_length=255, blank=True, null=True)
    not_soft_deleted_fk = models.ForeignKey(
        MyNotSoftDeletedModel, on_delete=models.CASCADE)



from django.db import models
from django.db.models import F


class Versioned(models.Model):
    autoincrement_version = True

    # Strict increment leads to problems in SimpleHistory (disable by default)
    strict_autoincrement_version = False

    version = models.IntegerField(editable=False)

    def save(self, *args, **kwargs):  # noqa: pylint=arguments-differ
        if not self.version:
            self.version = 1
        else:
            if self.autoincrement_version and self.pk:
                self.version = F('version') + 1 \
                    if self.strict_autoincrement_version \
                    else self.version + 1

        super().save(*args, **kwargs)

    def optimistic_concurrency_update(self, **kwargs):
        """
        Safe optimistic concurrent update. If the object was not modified
        since we fetched it than the object is updated and function will
        return `True`. If it was modified than the function will
        return `False` and the object will not be updated.

        NOTE 1: In an environment with a lot of concurrent updates
        this approach might be wasteful.

        NOTE 2: This approach does not protect from modifications made
        to the object outside this function. If you have other tasks
        that modify the data directly (e.g use `save()` directly)
        you need to make sure they use the version as well.

        Example:

            class Account(Versioned):
                balance = models.IntegerField(default=100)

                def withdraw(self, amount):
                    if self.balance < amount:
                        raise errors.RuntimeError()

                    result = self.balance - amount
                    return self.optimistic_concurrency_update(balance=balance)

            x = Account()
            x.withdraw(100)
            x.withdraw(100)

        :return: is the object updated
        :rtype: bool
        """
        # more detail here: https://medium.com/@hakibenita/how-to-manage-concurrency-in-django-models-b240fed4ee2  # noqa
        kwargs['version'] = self.version + 1
        model = type(self)
        updated = model.objects.filter(pk=self.pk, version=self.version) \
            .update(**kwargs)
        # TODO: trigger post_save event
        return updated > 0

    class Meta:
        abstract = True



from .uided import Uided, PUided
from .dated import Dated
from .versioned import Versioned
from .historized import Historized


class BaseHistorical(Uided, Dated, Versioned, Historized):  # type: ignore
    class Meta:
        ordering = ['created']
        abstract = True


class BasePHistorical(PUided, Dated, Versioned, Historized):  # type: ignore
    class Meta:
        ordering = ['created']
        abstract = True



from typing import Optional

from django.http import HttpRequest
from simple_history.models import HistoricalRecords


def get_current_request() -> Optional[HttpRequest]:
    return getattr(HistoricalRecords.thread, 'request', None)



from unittest.mock import patch

import pytest
from django.db.models import Q
from django.test import TestCase
from django.utils.crypto import get_random_string

from pik.core.shortcuts import (
    get_object_or_none, validate_and_create_object, validate_and_update_object,
    update_or_create_object)
from .factories import MySimpleModelFactory, TestNameModelFactory
from ..models import MySimpleModel


@pytest.fixture(params=[
    (MySimpleModel, MySimpleModelFactory),
])
def test_model(request):
    return request.param


def test_get_object_or_none(test_model):
    model, factory = test_model
    objs = factory.create_batch(10)

    obj = get_object_or_none(model, data=get_random_string())
    assert obj is None

    obj = get_object_or_none(model, data=objs[-1].data)
    assert obj is not None


def test_get_object_or_none_queryset(test_model):
    model, factory = test_model
    objs = factory.create_batch(10)

    obj = get_object_or_none(model.objects.filter(data=get_random_string()))
    assert obj is None

    obj = get_object_or_none(model.objects.filter(data=objs[-1].data))
    assert obj is not None


def test_get_object_or_none_manager(test_model):
    model, factory = test_model
    objs = factory.create_batch(10)

    obj = get_object_or_none(model.objects, data=get_random_string())
    assert obj is None

    obj = get_object_or_none(model.objects, data=objs[-1].data)
    assert obj is not None


def test_get_object_or_none_args(test_model):
    model, factory = test_model
    objs = factory.create_batch(10)

    obj = get_object_or_none(model, Q(data=get_random_string()))
    assert obj is None

    obj = get_object_or_none(model, Q(data=objs[-1].data))
    assert obj is not None


def test_validate_and_create_object(test_model):
    name1 = TestNameModelFactory.create()
    name2 = TestNameModelFactory.create()
    model, factory = test_model

    kwargs = {'data': get_random_string(), 'names': [name1, name2]}
    obj = validate_and_create_object(model, **kwargs)
    assert obj.pk
    assert name1 in obj.names.all()
    assert name2 in obj.names.all()


def test_validate_and_update_object__update(test_model):
    model, factory = test_model
    name1 = TestNameModelFactory.create()
    name2 = TestNameModelFactory.create()

    obj = factory.create()
    new_data = get_random_string()
    kwargs = {'data': new_data, 'names': [name1, name2]}

    res_obj, is_updated = validate_and_update_object(obj, **kwargs)
    assert res_obj.pk == obj.pk
    assert is_updated
    assert res_obj.data == new_data
    assert name1 in res_obj.names.all()
    assert name2 in res_obj.names.all()


def test_validate_and_update_object__no_update(test_model):
    model, factory = test_model
    name1 = TestNameModelFactory.create()
    name2 = TestNameModelFactory.create()

    obj = factory.create(names=(name1, name2))
    kwargs = {'data': obj.data, 'names': [name1, name2]}

    res_obj, is_updated = validate_and_update_object(obj, **kwargs)
    assert res_obj.pk == obj.pk
    assert not is_updated
    assert res_obj.data == obj.data
    assert name1 in res_obj.names.all()
    assert name2 in res_obj.names.all()


def test_update_or_create_object__create_without_search(test_model):
    model, _ = test_model
    new_data = get_random_string()
    name1 = TestNameModelFactory.create()
    name2 = TestNameModelFactory.create()
    kwargs = {'data':new_data, 'names': [name1, name2]}

    res_obj, is_updated, is_created = update_or_create_object(
        model, **kwargs)
    assert res_obj.pk
    assert not is_updated
    assert is_created
    assert name1 in res_obj.names.all()
    assert name2 in res_obj.names.all()


def test_update_or_create_object__create(test_model):
    model, factory = test_model
    obj = factory.create()
    new_data = get_random_string()
    name1 = TestNameModelFactory.create()
    name2 = TestNameModelFactory.create()
    kwargs = {'data':new_data, 'names': [name1, name2]}

    res_obj, is_updated, is_created = update_or_create_object(
        model, search_keys=dict(data=new_data), **kwargs)
    assert res_obj.pk != obj.pk
    assert not is_updated
    assert is_created
    assert name1 in res_obj.names.all()
    assert name2 in res_obj.names.all()


def test_update_or_create_object__no_update(test_model):
    model, factory = test_model
    name1 = TestNameModelFactory.create()
    name2 = TestNameModelFactory.create()

    obj = factory.create(names=(name1, name2))
    kwargs = {'data': obj.data, 'names': [name1, name2]}

    res_obj, is_updated, is_created = update_or_create_object(
        model, search_keys=dict(data=obj.data), **kwargs)
    assert res_obj.pk == obj.pk
    assert not is_updated
    assert not is_created
    assert name1 in res_obj.names.all()
    assert name2 in res_obj.names.all()


class TestNotCallM2MUpdate(TestCase):
    @staticmethod
    @patch('pik.core.shortcuts.model_objects._update_m2m_fields')
    def test_update_or_create_object(_update_m2m_fields):
        model, factory = MySimpleModel, MySimpleModelFactory
        obj = factory.create()
        new_data = get_random_string()

        res_obj, is_updated, is_created = update_or_create_object(
            model, search_keys=dict(data=new_data), data=new_data)
        assert res_obj.pk != obj.pk
        assert not is_updated
        assert is_created
        _update_m2m_fields.assert_not_called()



import pytest
from django.utils.crypto import get_random_string
from django.db.utils import IntegrityError

from ..models import Building, Apartment, Parking


def test_create_parent_without_child():
    addr1 = get_random_string()
    Building.objects.create(address=addr1)
    assert Building.objects.filter(address=addr1).count() == 1
    assert Apartment.objects.filter(address=addr1).count() == 0
    assert Parking.objects.filter(address=addr1).count() == 0


def test_every_child_will_create_new_parent():
    Parking.objects.create(address=get_random_string())
    Apartment.objects.create(address=get_random_string())
    Building.objects.create(address=get_random_string())
    assert Building.objects.all().count() == 3
    assert Apartment.objects.all().count() == 1
    assert Parking.objects.all().count() == 1


def test_child_will_create_new_parent_with_the_same_address_and_uid():
    addr1 = get_random_string()
    Parking.objects.create(address=addr1)
    assert Building.objects.all().count() == 1
    assert Apartment.objects.all().count() == 0
    assert Parking.objects.all().count() == 1

    building = Building.objects.all()[0]
    assert building.address == addr1
    parking = Parking.objects.all()[0]
    assert parking.address == addr1
    assert building.uid == parking.uid


def test_parent_address_unique_constraint():
    addr1 = get_random_string()
    Building.objects.create(address=addr1)
    with pytest.raises(IntegrityError):
        Parking.objects.create(address=addr1)
    assert Building.objects.filter(address=addr1).count() == 1
    assert Apartment.objects.filter(address=addr1).count() == 0
    assert Parking.objects.filter(address=addr1).count() == 0


def test_child_ptr_uid_the_same_as_parent():
    parking = Parking.objects.create(address=get_random_string())
    assert parking.uid == parking.building_ptr.uid



from inspect import getfullargspec
from functools import wraps
from typing import Any, Callable

from django.core.cache import caches


def cachedmethod(key: str, ttl: int = 5 * 60, cachename: str = 'default') \
        -> Callable:
    """
    Method result caching decorator

    Args:
        :param key: Template used to generate key with arguments
        :param ttl: cache ttl
        :param cachename: Name on django cache backend
        :return: value

    Examples:

        Keyword argument call:

            >>> @cachedmethod("keyword_{arg}")
            ... def calculate(arg: int) -> int:
            ...     return arg * 2
            >>> caches['default'].set('keyword_2', 10)
            True
            >>> calculate(arg=2)
            10

        Positional argument call:

            >>> @cachedmethod("positional_{a}")
            ... def calculate(a: int) -> int:
            ...     return a * 3
            >>> calculate(4)
            12
            >>> caches['default'].get('positional_4')
            12

        Default value call:

            >>> @cachedmethod("default_{b}_{c}")
            ... def calculate(a: int = 1, b: int = 3, c: int = 15) -> int:
            ...     return a + b * c
            >>> calculate()
            46
            >>> caches['default'].get('default_3_15')
            46
            >>> calculate(a=10000)
            46
            >>> calculate(10000)
            46
            >>> calculate(10000, b=3)
            46
            >>> calculate(a=10000, b=3)
            46
            >>> calculate(c=15, b=3, a=10000)
            46
            >>> calculate(**{'a': 1000})
            46
            >>> calculate(1000, a=2000)
            Traceback (most recent call last):
            ...
            TypeError: calculate() got multiple values for some argument

        Override ttl:

            >>> @cachedmethod("ttled_{a}_{b}_{c}", ttl=10*60)
            ... def calculate(a: int, b: int, c: int) -> int:
            ...     return a + b + c
            >>> calculate(1, 2, 3)
            6
            >>> calculate(1, 2, c=3)
            6

        Override cache:

            >>> @cachedmethod("othercache_{foo}", cachename="default")
            ... def calculate(foo: str) -> str:
            ...     return foo * 2
            >>> calculate('foo')
            'foofoo'

    """

    cache = caches[cachename]

    def wrapper(method: Callable) -> Callable:
        spec = getfullargspec(method)
        spec_keys = spec.defaults or ()
        defaults = dict(zip(spec.args[-len(spec_keys):], spec_keys))

        @wraps(method)
        def decorator(*args, **kwargs) -> Any:
            positional = dict(zip(spec.args, args))
            if set(positional) & set(kwargs):
                msg = '{name}() got multiple values for some argument'.format(
                    name=method.__name__)
                raise TypeError(msg)
            merged_kwargs = {**defaults, **positional, **kwargs}
            cachekey = key.format(**merged_kwargs)
            value = cache.get(cachekey)
            if value is None:
                value = method(*args, **kwargs)
                cache.set(cachekey, value, ttl)
            return value
        return decorator
    return wrapper



import pytest

from .factories import MyBasePHistoricalFactory, MyBaseHistoricalFactory
from ..models import MyBasePHistorical, MyBaseHistorical


@pytest.fixture(params=[
    (MyBasePHistorical, MyBasePHistoricalFactory),
    (MyBaseHistorical, MyBaseHistoricalFactory),
])
def historical_model(request):
    return request.param


def test_historical_protocol(historical_model):
    model, factory = historical_model

    obj1 = factory.create()
    obj2 = factory.create()
    obj_last = model.objects.last()
    obj_first = model.objects.first()
    assert obj_first.pk == obj1.pk
    assert obj_last.pk == obj2.pk
    assert obj_first.created < obj_last.created
    assert obj_first.updated < obj_last.updated
    assert obj_first.version == obj_last.version
    assert obj_first.uid != obj_last.uid


def test_historical_protocol_fields(historical_model):
    model, _ = historical_model
    fields = [f.name for f in model._meta.get_fields()]  # noqa: pylint=protected-access
    assert hasattr(model, 'history')
    assert 'uid' in fields
    assert 'version' in fields
    assert 'created' in fields
    assert 'updated' in fields


def test_historical_protocol_update(historical_model):
    _, factory = historical_model
    obj = factory.create()
    version1 = obj.version
    updated1 = obj.updated
    created1 = obj.created
    obj.save()
    version2 = obj.version
    updated2 = obj.updated
    created2 = obj.created
    obj.save()
    version3 = obj.version
    updated3 = obj.updated
    created3 = obj.created
    assert version1 < version2 < version3
    assert updated1 < updated2 < updated3
    assert created1 == created2 == created3


def test_historical_protocol_history(historical_model, settings):
    settings.SOFT_DELETE_SAFE_MODE = False
    model, factory = historical_model
    obj1 = factory.create()
    obj2 = factory.create()
    obj3 = factory.create()
    uid1, uid2, uid3 = obj1.uid, obj2.uid, obj3.uid

    obj1.save()
    obj2.save()
    obj3.save()

    obj3.delete()
    obj2.delete()
    obj1.delete()

    hist2 = [(x.uid, x.history_type, x.version) for x in model.history.all()]
    assert hist2 == [
        (uid1, '-', 2), (uid2, '-', 2), (uid3, '-', 2),
        (uid3, '~', 2), (uid2, '~', 2), (uid1, '~', 2),
        (uid3, '+', 1), (uid2, '+', 1), (uid1, '+', 1),
    ]



from django.db import models
from pik.core.models import BasePHistorical
from pik.core.models.fields import InheritPrimaryUidField


class Building(BasePHistorical):
    address = models.CharField(max_length=255, unique=True)

    def __str__(self):
        return f'{self.uid}'


class Apartment(Building):
    building_ptr = InheritPrimaryUidField(Building)


class Parking(Building):
    building_ptr = InheritPrimaryUidField(Building)


class BuildingPart(BasePHistorical):
    building = models.ForeignKey(Building, on_delete=models.CASCADE)

    def __str__(self):
        return f'{self.building_id} -> {self.uid}'


class Stairwell(BuildingPart):
    buildingpart_ptr = InheritPrimaryUidField(BuildingPart)


class Garret(BuildingPart):
    buildingpart_ptr = InheritPrimaryUidField(BuildingPart)



from .model_objects import (
    get_object_or_none, validate_and_create_object,
    validate_and_update_object, update_or_create_object)
from .request import get_current_request

__all__ = [
    'get_object_or_none',
    'validate_and_create_object',
    'validate_and_update_object',
    'update_or_create_object',
    'get_current_request',
]



from django.db import models


class InheritPrimaryUidField(models.OneToOneField):
    """
    If you need to have a parent union class for many child sub-models with
    the same UID for the parent and child instances using models inheritance.

    For example, you have one model A and you want to divide a modal instance
    to different sub-models. Like divide Users to InternalUsers and
    ExternalUsers.

    Code:

        from pik.core.models import PUided

        class Users(PUided):
            pass

        class InternalUsers(Users):
            users_ptr = InheritPrimaryUidField(Users)

        class ExternalUsers(Users):
            users_ptr = InheritPrimaryUidField(Users)

    Now you can use InternalUsers, ExternalUsers and Users for ForeignKey

    # How it works in migrations? #

    For example, we have a model:

        class Foo(models.Model):
            models.ForeignKey(Users, on_delete=models.CASCADE)

    And change it to:

        class Foo(models.Model):
            models.ForeignKey(InternalUsers, on_delete=models.CASCADE)

    Case I: We have a Foo object in DB with FK to ExternalUsers which
    not included in InternalUsers. When we run `migrate` command we will
    catch error: `django.db.utils.IntegrityError` insert or update
    on table "app_foo" violates foreign key constraint ...

    Case II: We don't have such objects. It this case everything will correct.

    """
    def __init__(self, to, **kwargs):
        kwargs.update(dict(
            on_delete=models.CASCADE, parent_link=True,
            primary_key=True, to_field='uid',
        ))
        super().__init__(to, **kwargs)

    def get_pk_value_on_save(self, instance):
        return instance.uid



import pytest
from django.contrib.auth import get_user_model
from django.db.utils import IntegrityError
from django.utils.timezone import now

from pik.core.models._collector_delete import DeleteNotSoftDeletedModel  # noqa: protected access
from test_core_models import models


def test_soft_deleted_version_bumps():
    obj = models.MySoftDeleteModel.objects.create(name='test')

    assert obj.version == 1

    obj.delete()

    assert obj.version == 2


def test_soft_deleted_updated_changed():
    obj = models.MySoftDeleteModel.objects.create(name='test')

    old_updated = obj.updated

    obj.delete()

    assert obj.updated > old_updated


def test_historized_soft_deleted():
    obj = models.MySoftDeleteModel.objects.create(name='test')
    history = obj.history.all()
    assert obj.history
    assert obj.history.count() == 1

    obj.save()
    assert obj.history.count() == 2

    obj.delete()
    assert obj.history.count() == 3
    assert history.count() == 3

    hist_obj1 = history.last()
    hist_obj2 = history[1]
    hist_obj3 = history.first()
    assert isinstance(hist_obj3.history_id, int)
    assert hist_obj3.pk == hist_obj3.history_id
    assert hist_obj3.history_change_reason is None
    assert hist_obj3.history_id > hist_obj1.history_id

    assert hist_obj1.history_type == '+'
    assert hist_obj2.history_type == '~'
    assert hist_obj3.history_type == '~'


def test_delete_soft_deleted_model_with_reverse_relation():
    obj = models.MySoftDeleteModel.objects.create(name='test')
    obj_with_rel = models.MyRelatedSoftDeletedModel.objects.create(
        name='super test', soft_deleted_fk=obj)

    old_obj_updated = obj.updated
    old_obj_with_rel_updated = obj_with_rel.updated

    obj.delete()
    obj_with_rel.refresh_from_db()

    assert obj.version == 2
    assert obj_with_rel.version == 2

    assert obj.deleted
    assert obj_with_rel.deleted

    assert obj.updated > old_obj_updated
    assert obj_with_rel.updated > old_obj_with_rel_updated


def test_restore_soft_deleted_model():
    obj = models.MySoftDeleteModel.objects.create(name='test')

    assert obj.version == 1

    obj.delete()

    updated_on_delete = obj.updated

    assert obj.deleted
    assert obj.version == 2

    obj.restore()

    assert not obj.deleted
    assert obj.version == 3
    assert obj.updated != updated_on_delete


def test_cascade_delete_with_fk_to_not_soft_deleted_model(settings):
    settings.SOFT_DELETE_SAFE_MODE = False
    not_soft_del_obj = models.MyNotSoftDeletedModel.objects.create(name='test')
    models.MySoftDeletedModelWithFK.objects.create(
        name='test soft', not_soft_deleted_fk=not_soft_del_obj)

    with pytest.raises(IntegrityError):
        not_soft_del_obj.delete()


def test_cascade_delete_with_fk_to_soft_deleted_model_failed(settings):
    settings.SOFT_DELETE_SAFE_MODE = True
    soft_del_obj = models.MySoftDeleteModel.objects.create(
        name='test soft')
    models.MyRelatedNotSoftDeletedModel.objects.create(
        name='test not soft', soft_deleted_fk=soft_del_obj)

    with pytest.raises(DeleteNotSoftDeletedModel):
        soft_del_obj.delete()


def test_cascade_delete_with_fk_to_soft_deleted_model_success(settings):
    settings.SOFT_DELETE_SAFE_MODE = False
    soft_del_obj = models.MySoftDeleteModel.objects.create(
        name='test soft')
    not_soft_del_obj = models.MyRelatedNotSoftDeletedModel.objects.create(
        name='test not soft', soft_deleted_fk=soft_del_obj)

    soft_del_obj.delete()

    with pytest.raises(models.MyRelatedNotSoftDeletedModel.DoesNotExist):
        not_soft_del_obj.refresh_from_db()


def test_cascade_soft_delete_with_already_deleted_model():
    soft_del_obj = models.MySoftDeleteModel.objects.create(
        name='test soft')
    soft_del_obj_with_rel = models.MyRelatedSoftDeletedModel.objects.create(
        name='test soft', soft_deleted_fk=soft_del_obj, deleted=now())

    old_updated = soft_del_obj_with_rel.updated

    soft_del_obj.delete()

    soft_del_obj_with_rel.refresh_from_db()

    assert old_updated == soft_del_obj_with_rel.updated


def test_not_soft_delete_without_safe_mode(settings):
    settings.SOFT_DELETE_SAFE_MODE = False
    obj = models.MyNotSoftDeletedModel.objects.create(
        name='test soft')

    obj.delete()


def test_hard_delete_with_safe_mode(settings):
    settings.SOFT_DELETE_SAFE_MODE = True
    obj = models.MySoftDeleteModel.objects.create(name="i'm gone")
    obj.hard_delete()

    with pytest.raises(models.MySoftDeleteModel.DoesNotExist):
        obj.refresh_from_db()


def test_hard_delete_without_safe_mode(settings):
    settings.SOFT_DELETE_SAFE_MODE = False
    obj = models.MySoftDeleteModel.objects.create(name="I'm gone too")
    obj.hard_delete()

    with pytest.raises(models.MySoftDeleteModel.DoesNotExist):
        obj.refresh_from_db()


def test_all_objects_is_deleted_filter(settings):
    settings.SOFT_DELETE_SAFE_MODE = True
    obj1 = models.MySoftDeleteModel.objects.create(name="obj 1")
    obj2 = models.MySoftDeleteModel.objects.create(name="obj 2")
    obj3 = models.MySoftDeleteModel.objects.create(name="obj 3")

    obj2.delete()
    obj3.delete()

    assert 3 == len(models.MySoftDeleteModel.all_objects.all())

    is_deleted_qs = models.MySoftDeleteModel.all_objects.is_deleted()
    assert 2 == len(is_deleted_qs)
    assert obj2 in is_deleted_qs
    assert obj3 in is_deleted_qs

    is_not_deleted_qs = models.MySoftDeleteModel.all_objects.is_not_deleted()
    assert 1 == len(is_not_deleted_qs)
    assert obj1 in is_not_deleted_qs


def test_deleted_model_from_exclude_list(settings):
    settings.SOFT_DELETE_SAFE_MODE = True
    settings.SOFT_DELETE_EXCLUDE = ['auth.User']

    User = get_user_model()
    u = User.objects.create(username='test_user')
    u.delete()

    with pytest.raises(User.DoesNotExist):
        u.refresh_from_db()



import pytest

from .factories import MyVersionedFactory, MyStrictVersionedFactory
from ..models import MyVersioned, MyStrictVersioned


@pytest.fixture(params=[
    (MyVersioned, MyVersionedFactory),
])
def versioned_model(request):
    return request.param


@pytest.fixture(params=[
    (MyStrictVersioned, MyStrictVersionedFactory),
])
def strict_versioned_model(request):
    return request.param


def test_versioned_protocol(versioned_model):
    model, factory = versioned_model
    obj = factory.create()
    version1 = obj.version
    assert isinstance(obj.version, int)
    assert obj.version > 0

    obj.save()
    version2 = obj.version
    assert version2 > version1


def test_strict_versioned_protocol(strict_versioned_model):
    model, factory = strict_versioned_model
    obj = factory.create()
    version1 = obj.version
    assert isinstance(obj.version, int)
    assert obj.version > 0

    obj.save()
    version2 = obj.version
    with pytest.raises(TypeError):
        assert version2 > version1

    obj.refresh_from_db()
    assert obj.version > version1


def test_optimistic_concurrency_update(versioned_model):
    model, factory = versioned_model
    obj = factory.create()
    version1 = obj.version
    is_updated = obj.optimistic_concurrency_update()
    obj.refresh_from_db()
    version2 = obj.version
    assert is_updated
    assert version1 < version2



from collections import defaultdict

from django.contrib.admin.utils import NestedObjects
from django.core.exceptions import FieldDoesNotExist
from django.db import models, router, transaction
from django.db.models import Q
from django.db.models.sql.where import WhereNode
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _
from ._collector_delete import Collector

assert Collector.delete


def _get_field_by_name(model, field):
    """
    Retrieve a field instance from a model by its name.
    """
    field_dict = {x.name: x for x in model._meta.get_fields()}  # noqa
    return field_dict[field]


def _has_field(model, field):
    try:
        model._meta.get_field(field)  # noqa
        return True
    except FieldDoesNotExist:
        return False


def _cascade_soft_delete(inst_or_qs, using, keep_parents=False):
    """
    Return collector instance that has marked ArchiveMixin instances for
    archive (i.e. update) instead of actual delete.
    Arguments:
        inst_or_qs (models.Model or models.QuerySet): the instance(s) that
            are to be deleted.
        using (db connection/router): the db to delete from.
        keep_parents (bool): defaults to False.  Determine if cascade is true.
    Returns:
        models.deletion.Collector: this is a standard Collector instance but
            the ArchiveMixin instances are in the fields for update list.
    """
    if not isinstance(inst_or_qs, models.QuerySet):
        instances = [inst_or_qs]
    else:
        instances = inst_or_qs

    deleted = now()

    # The collector will iteratively crawl the relationships and
    # create a list of models and instances that are connected to
    # this instance.
    collector = NestedObjects(using=using)
    collector.collect(instances, keep_parents=keep_parents)
    if collector.protected:
        raise models.ProtectedError("Delete protected", collector.protected)
    collector.sort()
    soft_delete_objs = collector.soft_delete_objs = defaultdict(set)

    for model, instances in list(collector.data.items()):
        # remove archive mixin models from the delete list and put
        # them in the update list.  If we do this, we can just call
        # the collector.delete method.
        if _has_field(model, 'deleted'):
            inst_list = [x for x in instances if x.deleted is None]
            deleted_on_field = _get_field_by_name(model, 'deleted')
            collector.add_field_update(deleted_on_field, deleted, inst_list)
            soft_delete_objs[model].update(inst_list)
            del collector.data[model]

    # If we use the NestedObjects collector instead models.deletion.Collector,
    # then the `collector.fast_deletes` will always be empty
    for i, q_set in enumerate(collector.fast_deletes):
        # make sure that we do archive on fast deletable models as
        # well.
        model = q_set.model
        if _has_field(model, 'deleted'):
            inst_list = [x for x in instances if x.deleted is None]
            deleted_on_field = _get_field_by_name(model, 'deleted')
            collector.add_field_update(deleted_on_field, deleted, inst_list)
            collector.fast_deletes[i] = q_set.none()

    return collector


def _delete_collected(collector):
    with transaction.atomic(using=collector.using, savepoint=False):
        result = collector.delete()
        for model, instances in collector.soft_delete_objs.items():
            if not model._meta.auto_created:  # noqa: pylint=protected-access
                for obj in instances:
                    obj.save()
    return result


class _BaseSoftDeletedQuerySet(models.QuerySet):
    def delete(self):
        # doing an update is the most efficient, but does not promise
        # that the cascade will happen. E.g.
        # return self.update(deleted_on=timezone.now())

        # from django source
        # https://github.com/django/django/blob/1.8.6/django/db/models/query.py
        # Line: #L516
        assert self.query.can_filter(), \
            "Cannot use 'limit' or 'offset' with delete."

        # iterating and deleting ensures that the cascade delete will
        # occur for each instance.
        collector = _cascade_soft_delete(self.all(), self.db)
        self._result_cache = None
        return _delete_collected(collector)

    delete.alters_data = True  # type: ignore
    delete.queryset_only = True  # type: ignore

    def hard_delete(self):
        return models.QuerySet.delete(self)

    hard_delete.alters_data = True  # type: ignore
    hard_delete.queryset_only = True  # type: ignore

    def restore(self):
        return self.update(**{'deleted': None})

    restore.alters_data = True  # type: ignore
    restore.queryset_only = True  # type: ignore

    def get_restore_or_create(self, **kwargs):
        obj, created = self.model.all_objects.get_or_create(**kwargs)
        if not created and obj.deleted:
            obj.restore()
        return obj

    def update_restore_or_create(self, **kwargs):
        obj, created = self.model.all_objects.update_or_create(**kwargs)
        if not created and obj.deleted:
            obj.restore()
        return obj


class _AllWhereNode(WhereNode):
    pass


class _SoftDeletedObjectsWhereNode(WhereNode):
    pass


class _SoftObjectsWhereNode(WhereNode):
    pass


class SoftObjectsQuerySet(_BaseSoftDeletedQuerySet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.query.where_class = _SoftObjectsWhereNode
        self.query.add_q(Q(deleted=None))


class SoftDeletedObjectsQuerySet(_BaseSoftDeletedQuerySet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.query.where_class = _SoftDeletedObjectsWhereNode
        self.query.add_q(~Q(deleted=None))


class AllObjectsQuerySet(_BaseSoftDeletedQuerySet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.query.where_class = _AllWhereNode

    def is_deleted(self):
        return self.filter(deleted__isnull=False)

    def is_not_deleted(self):
        return self.filter(deleted__isnull=True)


class SoftDeleted(models.Model):
    """
    Soft deletable model. Inspired by:
    https://lucasroesler.com/2017/04/delete-or-not-to-delete/

    If you want to use SoftDeleted with unique constraint there are
    a problem because NULL != NULL.
    You can use workaround with `UniqueConstraint in django>=2.2
    https://docs.djangoproject.com/en/2.2/ref/models/constraints/#django.db.models.UniqueConstraint
    """
    deleted = models.DateTimeField(
        editable=False, null=True, blank=True, verbose_name=_('Deleted')
    )

    objects = SoftObjectsQuerySet.as_manager()
    deleted_objects = SoftDeletedObjectsQuerySet.as_manager()
    all_objects = AllObjectsQuerySet.as_manager()

    def delete(self, using=None, keep_parents=False):
        using = using or router.db_for_write(self.__class__, instance=self)

        assert self._get_pk_val() is not None, \
            "%s object can't be deleted because its %s attribute " \
            "is set to None." % (self._meta.object_name, self._meta.pk.attname)

        if self.deleted:
            return 0, {}  # short-circuit here to prevent lots of nesting

        collector = _cascade_soft_delete(self, using, keep_parents)
        return _delete_collected(collector)

    delete.alters_data = True  # type: ignore

    def hard_delete(self, *args, **kwargs):
        return models.Model.delete(self, *args, **kwargs)

    def restore(self):
        self.deleted = None
        self.save()

    class Meta:
        abstract = True
